# core/config.py

"""
Handle Config init, dependencies check, logging setup,
files and folders structures, config settings parameters init
"""

from pathlib import Path
import logging
import os
import tomllib
import subprocess

from rcdl.core.file_io import write_txt


class Config:
    """Global app var/parameters"""

    # paths
    APP_NAME = "rcdl"

    BASE_DIR = Path(os.environ.get("RCDL_BASE_DIR", Path.home() / "Videos/rcdl"))

    CACHE_DIR = BASE_DIR / ".cache"
    DB_PATH = CACHE_DIR / "cdl.db"
    LOG_FILE = CACHE_DIR / "cdl.log"
    CREATORS_FILE = CACHE_DIR / "creators.txt"
    DISCOVER_DIR = CACHE_DIR / "discover"
    CONFIG_FILE = CACHE_DIR / "config.toml"

    DEBUG = False
    DRY_RUN = False

    # api settings
    POST_PER_PAGE: int = 50
    DEFAULT_MAX_PAGE: int = 10
    MAX_FAIL_COUNT: int = 7
    TIMEOUT: int = 10

    # fuse settings
    MAX_WIDTH: int = 1920
    MAX_HEIGHT: int = 1080
    FPS: int = 30
    PRESET: str = "veryfast"
    THREADS: int = 0

    HANDBRAKE_RUN_CMD = "HandBrakeCLI"

    CHECKSUM_RETRY = 2

    @classmethod
    def ensure_dirs(cls):
        """Ensure directory exist"""
        cls.CACHE_DIR.mkdir(parents=True, exist_ok=True)
        cls.DISCOVER_DIR.mkdir(exist_ok=True)

    @classmethod
    def ensure_files(cls):
        """Ensure file exist, populate default if necessary"""
        files = [cls.DB_PATH, cls.CREATORS_FILE, cls.CONFIG_FILE]
        for file in files:
            if not file.exists():
                file.touch()
                logging.info("Created file %s", file)
                if file == cls.CREATORS_FILE:
                    write_txt(cls.CREATORS_FILE, DEFAULT_CREATORS, mode="w")
                if file == cls.CONFIG_FILE:
                    write_txt(cls.CONFIG_FILE, DEFAULT_CONFIG, mode="w")

    @classmethod
    def creator_folder(cls, creator_id: str) -> Path:
        """Return creator folder path base on user/creator_id"""
        folder = cls.BASE_DIR / creator_id
        folder.mkdir(exist_ok=True)
        return folder

    @classmethod
    def cache_file(cls, filename: str, ext: str = ".json") -> Path:
        """Return filepath of a file in the .cache/ folder"""
        file_name = filename + ext
        file = cls.CACHE_DIR / file_name
        return file

    @classmethod
    def set_debug(cls, debug: bool):
        """Set class variable DEBUG"""
        cls.DEBUG = debug

    @classmethod
    def set_dry_run(cls, dry_run: bool):
        """Set class variable DRY_RUN"""
        cls.DRY_RUN = dry_run

    @classmethod
    def load_config(cls):
        """Load config.toml and set class var with value in config.toml"""
        with open(cls.CONFIG_FILE, "rb") as f:
            data = tomllib.load(f)
        app = data.get("app", {})
        cls.DEFAULT_MAX_PAGE = app.get("default_max_page", cls.DEFAULT_MAX_PAGE)
        cls.MAX_FAIL_COUNT = app.get("max_fail_count", cls.MAX_FAIL_COUNT)
        cls.TIMEOUT = app.get("timeout", cls.TIMEOUT)
        cls.CHECKSUM_RETRY = app.get("checksum_retry", cls.CHECKSUM_RETRY)

        video = data.get("video", {})
        cls.MAX_WIDTH = video.get("max_width", cls.MAX_WIDTH)
        cls.MAX_HEIGHT = video.get("max_height", cls.MAX_HEIGHT)
        cls.FPS = video.get("fps", cls.FPS)
        cls.PRESET = video.get("preset", cls.PRESET)
        cls.THREADS = video.get("threads", cls.THREADS)

        paths = data.get("paths", {})
        if "base_dir" in paths:
            cls.BASE_DIR = Path(
                os.environ.get("RCDL_BASE_DIR", os.path.expanduser(paths["base_dir"]))
            )
            cls.CACHE_DIR = cls.BASE_DIR / ".cache"
        if "handbrake_run_cmd" in paths:
            cls.HANDBRAKE_RUN_CMD = paths.get("handbrake_run_cmd")


def setup_logging(log_file: Path, level: int = 0):
    """Setup logging for rcdl"""
    logger = logging.getLogger()
    logger.setLevel(level)
    logger.handlers.clear()  # avoid double handlers if called multiple times

    # loggin format & file handler
    file_handler = logging.FileHandler(log_file, encoding="utf-8", mode="a")
    file_handler.setFormatter(
        logging.Formatter(
            "{asctime} - {levelname} - {message}",
            style="{",
            datefmt="%Y-%m-%d %H:%M:%S",
        )
    )
    logger.addHandler(file_handler)

    # log library warning/errors
    stream = logging.StreamHandler()
    stream.setLevel(logging.ERROR)  # only show warnings/errors from libraries
    logger.addHandler(stream)


def check_dependencies():
    """Check external program version against last tested working version"""
    for prgrm, info in DEPENDENCIES_TEST_VERSION.items():
        try:
            result = subprocess.run(
                info["cmd"],
                capture_output=True,
                text=True,
                shell=True,
                check=False,
            )
            version = result.stdout.strip()

            if version != info["version"]:
                print(
                    f"Last tested version for {prgrm}:"
                    f" {info['version']} -> yours: {version}"
                )
                if version == "":
                    print(f"{prgrm} is not installed.")
                print(f"Check {prgrm} is installed if your version is empty.")
        except (OSError, subprocess.SubprocessError) as e:
            print(
                f"Failed to check {prgrm} version due to: {e}\nCheck {prgrm} is installed."
            )


DEPENDENCIES_TEST_VERSION = {
    "yt-dlp": {"cmd": "yt-dlp --version", "version": "2025.12.08"},
    "aria2c": {
        "cmd": "aria2c -v | head -n 1",
        "version": "aria2 version 1.37.0",
    },
    "ffmpeg": {
        "cmd": 'ffmpeg -version | sed -n "s/ffmpeg version \\([-0-9.]*\\).*/\\1/p;"',
        "version": "7.1.1-1",
    },
    "handbrake": {
        "cmd": Config.HANDBRAKE_RUN_CMD
        + ' --version 2>&1 | sed -n "s/HandBrake \\([0-9.]*\\).*/\\1/p"',
        "version": "1.9.2",
    },
}


# default creators
DEFAULT_CREATORS = ["boixd/onlyfans"]

# default config params
DEFAULT_CONFIG: str = """\
[app]
default_max_page = 10
max_fail_count = 7
timeout = 10
checksum_retry = 2

[fuse]
max_width = 1920
max_height = 1080
fps = 30
preset = "veryfast"
threads = 0

[paths]
base_dir = "~/Videos/rcdl"
handbrake_run_cmd = "HandBrakeCLI"
"""



# interface/ui.py

import logging
import click
from rich.console import Console, Group
from rich.table import Table
from rich.progress import (
    Progress,
    SpinnerColumn,
    BarColumn,
    TextColumn,
    TimeRemainingColumn,
    TimeElapsedColumn,
)
from rich import box
from rich.live import Live
from rich.text import Text
from rcdl.core.models import Creator, FusedStatus, Status


class NestedProgress:
    def __init__(self, console: Console):
        self.console = console
        self.global_progress: Progress | None = None
        self.current_progress: Progress | None = None
        self.live: Live | None = None

        self.total_task: int | None = None
        self.current_task: int | None = None
        self.status_text = Text("", style="cyan")
        self.current_label = ""

    def start(
        self, *, total: int, total_label: str = "Total", current_label: str = "Current"
    ):
        # Global progress (elapsed only)
        self.global_progress = Progress(
            SpinnerColumn(),
            TextColumn("[bold cyan]{task.description}"),
            BarColumn(),
            TextColumn("{task.completed}/{task.total}"),
            TimeElapsedColumn(),
            console=self.console,
            transient=False,
        )

        # Current task progress (ETA included)
        self.current_progress = Progress(
            SpinnerColumn(),
            TextColumn("[bold cyan]{task.description}"),
            BarColumn(),
            TextColumn("{task.completed}/{task.total}"),
            TimeRemainingColumn(),
            console=self.console,
            transient=False,
        )

        self.current_label = current_label
        group = Group(self.global_progress, self.current_progress, self.status_text)
        self.live = Live(group, console=self.console)
        self.live.__enter__()

        self.total_task = self.global_progress.add_task(total_label, total=total)
        self.current_task = self.current_progress.add_task(
            current_label, total=1, visible=False
        )

    # total task helpers
    def advance_total(self, step: int = 1):
        if self.global_progress and self.total_task is not None:
            self.global_progress.advance(self.total_task, step)  # type: ignore

    def reset_current(self):
        if not self.current_progress or self.current_task is None:
            return

        self.current_progress.remove_task(self.current_task)  # type: ignore

        self.current_task = self.current_progress.add_task(
            self.current_label,
            total=1,
            visible=False,
        )

    def start_current(self, description: str, total: int | None = None):
        if not self.current_progress or self.current_task is None:
            return

        self.reset_current()

        self.current_progress.update(
            self.current_task,  # type: ignore
            description=description,
            total=total or 1,
            completed=0,
            visible=True,
        )

    def advance_current(self, step: int | float = 1):
        if self.current_progress and self.current_task is not None:
            self.current_progress.advance(self.current_task, step)  # type: ignore

    def finish_current(self):
        if self.current_progress and self.current_task is not None:
            self.current_progress.update(self.current_task, visible=False)  # type: ignore

    # Status line
    def set_status(self, cyan: str, green: str = ""):
        self.status_text.plain = ""
        self.status_text.append(cyan, style="cyan")
        if green:
            self.status_text.append(green, style="green")

    # Close
    def close(self):
        if self.live:
            self.live.__exit__(None, None, None)
            self.live = None


class UI:
    console = Console()
    logger = logging.getLogger()

    _video_progress_text: Text | None = None
    _concat_progress_text: Text | None = None
    _live: Live | None = None
    _generic_text: Text | None = None

    @staticmethod
    def _log_to_file(log_level, msg: str):
        log_level(msg)

    @classmethod
    def success(cls, msg: str):
        """Print success msg"""
        cls.console.print(f"[green]{msg}[/]")

    @classmethod
    def info(cls, msg: str):
        """Print & log info msg"""
        cls.console.print(msg)
        cls._log_to_file(cls.logger.info, msg)

    @classmethod
    def debug(cls, msg: str):
        """Log debug msg"""
        # cls.console.print(f"[dim]{msg}[/]")
        cls._log_to_file(cls.logger.debug, msg)

    @classmethod
    def warning(cls, msg: str):
        """Print & log warning msg"""
        cls.console.print(f"[yellow]{msg}[/]")
        cls._log_to_file(cls.logger.debug, msg)

    @classmethod
    def error(cls, msg: str):
        """Print & log error msg"""
        cls.console.print(f"[red]{msg}[/]")
        cls._log_to_file(cls.logger.debug, msg)

    @classmethod
    def table_creators(cls, creators: list[Creator]):
        """Print to cli a table with all creators in creators.txt. Format is Creator ID | Service"""
        table = Table(title="Creators", box=box.MINIMAL, show_lines=True)
        table.add_column("Creators ID")
        table.add_column("Service")
        for creator in creators:
            table.add_row(creator.id, creator.service)
        cls.console.print(table)

    @classmethod
    def progress_posts_fetcher(cls, max_pages: int):
        progress = Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            console=cls.console,
            transient=False,  # remove progress bar after finish
        )
        return progress


def print_db_info(info: dict):
    click.echo("--- TABLES ---")
    click.echo("Posts:")
    click.echo(f"\t{info['posts']} total")
    click.echo("FusedMedias:")
    for status in FusedStatus:
        click.echo(f"\t{info['fuses'][status]} {status}")
    click.echo("Medias:")
    for status in Status:
        click.echo(f"\t{info['medias'][status]} {status}")



# core/opti.py

"""
Optimize media to reduce disk storage utilisation
"""

import os

from rcdl.core.config import Config
from rcdl.core.models import Status, Media
from rcdl.core.db import DB
from rcdl.core.downloader_subprocess import handbrake_optimized
from rcdl.interface.ui import UI, NestedProgress
from rcdl.utils import get_media_metadata, get_date_now


def update_db(media: Media, user: str, result: int):
    """Update DB if optimisation succesfful with new file_size, etc..."""
    if result == 0:
        path = os.path.join(Config.creator_folder(user), media.file_path)
        _, file_size, checksum = get_media_metadata(path)
        media.status = Status.OPTIMIZED
        media.checksum = checksum
        media.created_at = get_date_now()
        media.file_size = file_size

        with DB() as db:
            db.update_media(media)


def optimize():
    """Optimize all medias in DB with DOWNLOADED
    status that are not part of a fuse group"""
    # get all video to opti
    with DB() as db:
        medias = db.query_media_by_status(Status.DOWNLOADED)
        if Config.DEBUG:
            medias.extend(db.query_media_by_status(Status.OPTIMIZED))

    # progress
    progress = NestedProgress(UI.console)
    progress.start(
        total=len(medias),
        total_label="Optimizing videos",
        current_label="Current video",
    )

    for media in medias:
        # check media is not in a fuse group
        with DB() as db:
            fuse = db.query_fuses_by_id(media.post_id)
        if fuse is not None:
            progress.advance_total()
            continue

        # get post info
        with DB() as db:
            post = db.query_post_by_id(media.post_id)
        if post is None:
            UI.error(f"Could not match media {media.post_id} to a post by id")
            progress.advance_total()
            continue

        result = handbrake_optimized(media, post.user, progress)

        folder_path = Config.creator_folder(post.user)
        video_path = os.path.join(folder_path, media.file_path)
        output_path = video_path + ".opti.mp4"

        if result == 0:
            try:
                os.replace(output_path, video_path)
                update_db(media, post.user, result)
            except FileNotFoundError as e:
                UI.error(
                    f"FileNotFoundError: Could not replace {video_path} "
                    f"with {output_path} due to: {e}"
                )
            except PermissionError as e:
                UI.error(
                    f"PermissionError: Could not replace {video_path} "
                    f"with {output_path} due to: {e}"
                )
            except OSError as e:
                UI.error(
                    f"OSError: Failed to replace {video_path} with {output_path} due to: {e}"
                )
            finally:
                progress.advance_total()
    progress.close()



# __main__.py

"""
__main__: entry in the programm
Setup logging, create files/folders structures, check dependencies,
init Config global variables, init database,
Then call cli group
"""

import logging

from rcdl.core.config import Config, setup_logging, check_dependencies

# setup file structure
Config.ensure_dirs()
Config.ensure_files()

# load config file settings
Config.load_config()

# setup logging
setup_logging(Config.LOG_FILE, level=0)

# check dependencies
check_dependencies()

logging.info("--- INIT ---")
logging.info("Logger initialized")

# init database
from rcdl.core.db import DB  # noqa: E402

db = DB()
db.init_database()
db.close()

from rcdl.interface.cli import cli  # noqa: E402, F401



# core/downloader.py

"""
Handle post, media download to website
"""

import logging
import os
import json

import requests

from rcdl.interface.ui import UI, NestedProgress
from rcdl.core import parser
from rcdl.core import adapters
from rcdl.core.api import URL
from rcdl.core.config import Config
from rcdl.core.models import (
    Creator,
    Status,
    Media,
    Post,
    FusedMedia,
    FusedStatus,
)
from rcdl.core.db import DB
from rcdl.core.downloader_subprocess import ytdlp_subprocess
from rcdl.core.file_io import write_json, load_json
from rcdl.utils import get_date_now, get_media_metadata


class PostsFetcher:
    """
    Fetch posts from api. Save as JSON. Handle multiple pages requests
    """

    def __init__(
        self, url: str, json_path: str, max_page: int = Config.DEFAULT_MAX_PAGE
    ):
        self.url = url
        self.json_path = json_path

        self.page = 0
        self.max_page = max_page

        self.status = 200

    def _request_page(self, url: str) -> requests.Response:
        """Request a single page and return json dict"""
        logging.info("RequestEngine url %s", url)
        headers = URL.get_headers()
        response = requests.get(url, headers=headers, timeout=Config.TIMEOUT)
        if response.status_code != 200:
            logging.warning("Failed request %s: %s", url, response.status_code)
        return response

    def request(self, params: dict | None = None):
        """Request multiple page of an url"""
        if params is None:
            params = {}

        with UI.progress_posts_fetcher(self.max_page) as progress:
            task = progress.add_task("Fetching posts", total=self.max_page)

            while self.status == 200 and self.page < self.max_page:
                o = self.page * Config.POST_PER_PAGE
                params["o"] = o
                url = URL.add_params(self.url, params)

                try:
                    # Dry run: not request acutally made
                    if Config.DRY_RUN:
                        logging.debug(
                            "DRY-RUN posts fetcher %s -> %s", url, self.json_path
                        )
                        self.page += 1
                        continue

                    response = self._request_page(url)
                    self.status = response.status_code

                    # if the programm crash while doing requests,
                    # previous requests are still saved and not overwritten.
                    if self.page > 0:
                        json_data = list(load_json(self.json_path))
                    else:
                        json_data = []

                    # for discover command, response json is in a
                    # different format and contains 'posts'
                    if self.status == 200:
                        if "posts" in response.json():
                            json_data.extend(response.json()["posts"])
                        else:
                            json_data.extend(response.json())

                        write_json(self.json_path, json_data, mode="w")

                    progress.update(
                        task,
                        advance=1,
                        description=(
                            f"Fetched {len(json_data)}"
                            f" posts (page {self.page + 1}/{self.max_page})"
                        ),
                    )
                except requests.RequestException as e:
                    logging.error(
                        "Failed to request %s (page: %s) deu to: %s", url, self.page, e
                    )
                except json.JSONDecodeError as e:
                    logging.error(
                        "Failed to decode JSON response of request %s due to: %s",
                        url,
                        e,
                    )
                finally:
                    self.page += 1


class MediaDownloader:
    """Handle downloading a list of media and update DB status"""

    def __init__(self):
        pass

    def _build_url(self, domain: str, url: str):
        """Return full url"""
        return URL.get_url_from_file(domain, url)

    def _build_full_path(self, user: str, media_path: str):
        """Return full path"""
        return os.path.join(Config.creator_folder(user), media_path)

    def _media_exist(self, full_path: str):
        """Check a file exist"""
        return os.path.exists(full_path)

    def _update_db(self, result: int, media: Media, full_path: str):
        """Update db information"""

        # video failed to download
        if result != 0:
            media.fail_count += 1
        else:
            duration, file_size, checksum = get_media_metadata(full_path)
            media.duration = duration
            media.status = Status.DOWNLOADED
            media.checksum = checksum
            media.created_at = get_date_now()
            media.file_size = file_size

        with DB() as db:
            db.update_media(media)

    def download(self, medias: list[Media], max_fail_count: int | None = None):
        """Download all medias in media with PENDING stats"""
        # init progress bar
        progress = NestedProgress(UI.console)
        progress.start(
            total=len(medias),
            total_label="Downloading videos",
            current_label="Current video",
        )

        max_try = Config.MAX_FAIL_COUNT
        if max_fail_count is not None:
            max_try = max_fail_count
        for media in medias:
            progress.start_current("Downloading", total=2)
            if media.fail_count > max_try:
                UI.warning(
                    f"Video skipped due to too many failed download attempt ({media.fail_count})"
                )
                progress.advance_total()
                continue

            # match post info from db with post_id to get user/creator_id
            with DB() as db:
                post = db.query_post_by_id(media.post_id)
            if post is None:
                UI.error(f"Could not match media post_id {media.post_id} with a post")
                progress.advance_total()
                continue

            # build full url and full path
            url = self._build_url(post.domain, media.url)
            full_path = self._build_full_path(post.user, media.file_path)

            # update progress bar info (video in download info)
            progress.set_status(f"{post.user}@({post.service}) -> ", media.file_path)

            # check video does not already exist
            if self._media_exist(full_path):
                UI.warning(
                    f"Video {url} @ {full_path} already exists. Possible DB problem"
                )
                self._update_db(0, media, full_path)
                progress.advance_total()
                continue

            # dry run: no actual download, skippe rest of fn
            if Config.DRY_RUN:
                UI.debug(f"(dry-run) dl  {post.user}@{full_path} from {url}")
                progress.advance_total()
                continue

            result = ytdlp_subprocess(url, full_path)
            self._update_db(result, media, full_path)
            progress.advance_total()
        progress.close()


def fetch_posts_by_tag(tag: str, max_page: int = Config.DEFAULT_MAX_PAGE) -> dict:
    """Helper function to get all posts from a search results"""
    url = URL.get_posts_page_url_wo_param()
    path = Config.cache_file(tag)
    pf = PostsFetcher(url, str(path), max_page=max_page)
    pf.request(params={"tag": tag})

    return load_json(path)


def fetch_posts_by_creator(creator: Creator) -> dict:
    """Helper function to get all posts from a creator"""
    url = URL.get_creator_post_wo_param(creator)
    path = Config.cache_file(f"{creator.id}_{creator.service}")
    pf = PostsFetcher(url, str(path))
    pf.request()

    return load_json(path)


def get_fuses_from_post(posts: list[Post]) -> list[FusedMedia]:
    """Update data on fuses database table for video to be fused"""
    fuses: list[FusedMedia] = []
    for post in posts:
        json_post = json.loads(post.raw_json)
        total_parts = len(parser.extract_video_urls(json_post))
        if total_parts > 1:
            fuses.append(
                FusedMedia(
                    id=post.id,
                    duration=0,
                    total_parts=total_parts,
                    status=FusedStatus.PENDING,
                    checksum="",
                    file_path=parser.get_filename_fuse(post),
                    created_at="",
                    updated_at="",
                    file_size=0,
                    fail_count=0,
                )
            )
    return fuses


def refresh_creators_videos():
    """
    For each creator:
        - get posts with videos & update posts DB
        - extract all medias & update medias DB
        - extract fuses group & update fuses DB
    """
    creators = parser.get_creators()
    for creator in creators:
        UI.info(f"Creator {creator.id} from {creator.service}")

        # request all posts by creator
        fetch_posts_by_creator(creator)

        # only keep posts with video url (mp4, m4v, ...)
        posts_with_videos = parser.filter_posts_with_videos_from_json(
            str(Config.cache_file(f"{creator.id}_{creator.service}"))
        )

        # convert all json dict into Post model
        posts = adapters.json_posts_to_posts(posts_with_videos)

        # insert posts in db
        with DB() as db:
            db.insert_posts(posts)

        # find all multiple part videos and update db
        fuses = get_fuses_from_post(posts)
        with DB() as db:
            db.insert_fused_media(fuses)

        # convert all posts into videos
        medias = []
        for post in posts:
            medias.extend(adapters.post_to_videos(post))

        # insert videos in db
        with DB() as db:
            db.insert_medias(medias)


def download_videos_to_be_dl(max_fail_count: int | None):
    """
    Download all media with PENDING status in DB
    """
    with DB() as db:
        medias = db.query_media_by_status(Status.PENDING)

    media_downloader = MediaDownloader()
    media_downloader.download(medias, max_fail_count=max_fail_count)



# core/parser.py

"""Handle function to parse post and files"""

import logging
from pathvalidate import sanitize_filename

from rcdl.interface.ui import UI
from rcdl.core.models import Media, Creator, Post, CreatorStatus
from rcdl.core.file_io import load_json, load_txt, write_txt
from rcdl.core.config import Config


COOMER_PAYSITES = ["onlyfans", "fansly", "candfans"]
KEMONO_PAYSITES = [
    "patreon",
    "fanbox",
    "fantia",
    "boosty",
    "gumroad",
    "subscribestar",
    "dlsite",
]


def get_domain(arg: str | dict | Media) -> str:
    """From a service get the domain (coomer or kemono)
    Input is either: service(str), post(dict), video(models.Video)
    """

    def _service(service: str) -> str:
        if service in COOMER_PAYSITES:
            return "coomer"
        if service in KEMONO_PAYSITES:
            return "kemono"
        logging.error("Service %s not associated to any domain", service)
        return ""

    if isinstance(arg, dict):
        return _service(arg["service"])
    if isinstance(arg, Media):
        return _service(arg.service)

    return _service(arg)


def get_title(post: Post) -> str:
    """From a Post Model return the title"""
    title = post.title
    if title == "":
        title = post.substring
    if title == "":
        title = post.id
    return sanitize_filename(title)


def get_title_json(post: dict) -> str:
    """Extract title from a post(dict)"""
    title = post["title"]
    if title == "":
        title = post["substring"]
    if title == "":
        title = post["id"]
    return sanitize_filename(title)


def get_date(post: dict) -> str:
    """Extract date from a post(dict)"""
    if "published" in post:
        date = post["published"][0:10]
    elif "added" in post:
        date = post["added"][0:10]
    else:
        logging.error("Could not extract date from %s", post["id"])
        date = "NA"
    return date


def get_part(post: dict, url: str) -> int:
    """
    For posts containing multiple video url. Each url is considered a part,
    so all videos from the same posts will simply have a different part number
    """
    urls = extract_video_urls(post)
    part = 0
    if len(urls) == 1:
        return 0

    for u in urls:
        if u == url:
            return part
        part += 1

    logging.error(
        "Could not extract part number for post id %s with url %s", post["id"], url
    )
    return -1


def get_filename(post: dict, url: str) -> str:
    """Get filename from pst dict and url"""
    title = get_title_json(post)
    date = get_date(post)
    part = get_part(post, url)
    file_title = f"{date}_{title}".replace("'", " ").replace('"', "")
    filename = f"{file_title}_p{part}.mp4"
    return filename


def get_filename_fuse(post: Post) -> str:
    """Get filename for fuse output from Post Model
    Fuse output has 'X' as part number"""
    title = get_title(post)
    date = post.published[0:10]
    part = "X"
    file_title = f"{date}_{title}".replace("'", " ").replace('"', "")
    filename = f"{file_title}_p{part}.mp4"
    return filename


def extract_video_urls(post: dict) -> list:
    """Extract all videos urls from a dict post"""
    video_extensions = (".mp4", ".webm", ".mov", ".avi", ".mkv", ".flv", ".wmv", ".m4v")
    urls = set()

    # Check main file
    if post["file"]:
        if post["file"]["path"]:
            path = post["file"]["path"]
            if path.endswith(video_extensions):
                urls.add(f"{path}")

    if post["attachments"]:
        attachments = post["attachments"]
        for attachment in attachments:
            if attachment["path"]:
                if attachment["path"].endswith(video_extensions):
                    urls.add(f"{attachment['path']}")

    return list(urls)


def filter_posts_with_videos_from_list(data: list[dict]) -> list[dict]:
    """Return posts with video url from a json with a list of posts"""

    posts_with_videos = []
    for post in data:
        if len(extract_video_urls(post)) > 0:
            posts_with_videos.append(post)
    return posts_with_videos


def filter_posts_with_videos_from_json(path: str) -> list:
    """Return posts with video url from a json with a list of posts"""
    posts = load_json(path)

    posts_with_videos = []
    for post in posts:
        if len(extract_video_urls(post)) > 0:
            posts_with_videos.append(post)
    return posts_with_videos


def valid_service(service: str) -> bool:
    """Check if a service is valid (within list of DOMAIN services)"""
    if service in COOMER_PAYSITES:
        return True
    if service in KEMONO_PAYSITES:
        return True
    return False


def _default_creator(_id: str, service: str, domain: str):
    return Creator(
        id=_id,
        service=service,
        domain=domain,
        name="",
        indexed="",
        updated="",
        favorited=1,
        status=CreatorStatus.NA,
        max_date="",
        max_posts=1,
        max_size=1,
        min_date="",
    )


def get_creator_from_line(line: str) -> Creator | None:
    """
    Convert a line into a Creator model
    arg: line -> 'service/creator'
    This is the format of creators.txt
    """

    parts = line.split("/")
    if valid_service(parts[0].strip()):
        return _default_creator(
            parts[1].strip(), parts[0].strip(), get_domain(parts[0].strip())
        )
    if valid_service(parts[1].strip()):
        return _default_creator(
            parts[0].strip(), parts[1].strip(), get_domain(parts[1].strip())
        )

    UI.error(
        f"Creator file not valid: {line} can not be interpreted."
        f" Format is: 'service/creator_id'"
    )
    return None


def get_creators() -> list[Creator]:
    """
    Load creators.txt and return a list of models.Creator
    """
    lines = load_txt(Config.CREATORS_FILE)
    creators = []
    for line in lines:
        creator = get_creator_from_line(line)
        if creator is None:
            continue
        creators.append(creator)
    if len(creators) < 1:
        UI.error(f"Could not find any creators. Check {Config.CREATORS_FILE}")
    return creators


def get_creators_from_posts(posts: list[dict]) -> list[Creator]:
    """Extract a list of Creators model form a list of dict posts"""
    creators = []
    seen = set()

    for post in posts:
        key = (post["user"], post["service"], "coomer")
        if key in seen:
            continue

        seen.add(key)
        creators.append(_default_creator(post["user"], post["service"], "coomer"))
    return creators


def parse_creator_input(value: str) -> tuple[str | None, str]:
    """Parse user input in cli to extract creator id & service"""
    value = value.strip()

    # url
    if "://" in value:
        parts = value.replace("https://", "").strip().split("/")
        logging.info(
            "From %s extracte service %s and creator %s", value, parts[1], parts[3]
        )
        return parts[1], parts[3]  # service, creator_id

    # creators.txt format
    if "/" in value:
        c = get_creator_from_line(value)
        if c is not None:
            logging.info(
                "From %s extracte service %s and creator %s",
                value,
                c.service,
                c.id,
            )
            return c.service, c.id

    logging.info("From %s extracted service None and creator %s", value, value)
    return None, value


def append_creator(creator: Creator):
    """Append a creator to the creators.txt file
    Creators.txt hold all creators used in refresh command"""
    line = f"{creator.service}/{creator.id}"
    lines = load_txt(Config.CREATORS_FILE)

    if line in lines:
        return
    lines.append(line)
    write_txt(Config.CREATORS_FILE, line, mode="a")



# core/downloader_subprocess.py

"""
Handle all subprocess call to external tool (yt-dlp, ffmpeg, ...)
"""

import subprocess
import logging
from pathlib import Path
import os


from rcdl.interface.ui import UI, NestedProgress
from rcdl.core import parser
from rcdl.core.models import Media, Post
from rcdl.core.config import Config
from rcdl.utils import bytes_to_str


def ytdlp_clear_cache():
    """Clear yt-dlp cache"""
    cmd = ["yt-dlp", "--rm-cache-dir"]
    subprocess.run(cmd, check=False)


def kill_aria2c():
    """Kill all aria2c process"""
    cmd = ["pkill", "-f", "aria2c"]
    subprocess.run(cmd, check=False)


def ytdlp_subprocess(
    url: str,
    filepath: Path | str,
):
    """Call yt-dlp in a subprocess to download a video"""

    cmd = [
        "yt-dlp",
        "-q",
        "--progress",
        url,
        "-o",
        filepath,
        "--external-downloader",
        "aria2c",
    ]

    logging.info("CMD: %s", " ".join(cmd))

    result = subprocess.run(cmd, capture_output=True, text=True, check=False)
    if result.returncode != 0:
        logging.error("yt-dlp failed to dl vid: %s", result.stderr)

    return result.returncode


def ffprobe_get_duration(path: Path) -> int | None:
    """Get duration of a video in seconds with ffprobe
    Return an int or None if command failed"""
    cmd = [
        "ffprobe",
        "-v",
        "error",
        "-show_entries",
        "format=duration",
        "-of",
        "default=noprint_wrappers=1:nokey=1",
        str(path),
    ]
    try:
        result = subprocess.run(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
            text=True,
            check=True,
        )
        return int(float(result.stdout.strip()))
    except subprocess.CalledProcessError as e:
        UI.error(f"Failed to use ffprobe on {path} due to {e}")
        return None
    except (AttributeError, ValueError, OverflowError) as e:
        UI.error(f"Failed to parse duration result of {path} due to {e}")
        return None


def get_max_width_height(medias: list[Media], post: Post) -> tuple[int, int]:
    """Get width and height of all media in list. Return max within video found and config"""

    cmd = [
        "ffprobe",
        "-v",
        "error",
        "-select_streams",
        "v:0",
        "-show_entries",
        "stream=width,height",
        "-of",
        "csv=p=0",
    ]
    width = 0
    height = 0
    max_width = 1920
    max_height = 1080
    for m in medias:
        path = os.path.join(Config.creator_folder(post.user), m.file_path)
        full_cmd = cmd + [path]

        try:
            result = subprocess.run(
                full_cmd, capture_output=True, text=True, check=True
            )
            w_str, h_str = result.stdout.strip().split(",")

            width = min(int(w_str), max_width)
            height = min(int(h_str), max_height)
        except subprocess.CalledProcessError as e:
            UI.error(f"Fail to use ffprobe to get width, height on {path} due to {e}")
        except (AttributeError, ValueError, OverflowError) as e:
            UI.error(f"Failed to parse duration for {path} due to {e}")
    return (width, height)


def get_total_duration(medias: list[Media], post: Post) -> int:
    """Get total duration in ms of all medias in list"""

    def _get_duration(path: str) -> int:
        """Get video duration in ms"""

        cmd = [
            "ffprobe",
            "-v",
            "error",
            "-select_streams",
            "v:0",
            "-show_entries",
            "format=duration",
            "-of",
            "default=noprint_wrappers=1:nokey=1",
            path,
        ]

        result = subprocess.run(cmd, capture_output=True, text=True, check=True)
        return int(float(result.stdout.strip()) * 1000)

    duration = 0
    for m in medias:
        path = os.path.join(Config.creator_folder(post.user), m.file_path)
        duration += _get_duration(path)
    return duration


def ffmpeg_concat_build_command(medias: list[Media], post: Post) -> dict:
    """Build the ffmpeg concat command"""

    width, height = get_max_width_height(medias, post)
    logging.info("Found (%s, %s) (width, height) for this group.", width, height)
    if width == 0:
        width = Config.MAX_WIDTH
    if height == 0:
        height = Config.MAX_HEIGHT

    # output path
    output_filename = parser.get_filename_fuse(post)
    output_path = os.path.join(Config.creator_folder(post.user), output_filename)

    # build cmd
    cmd = ["ffmpeg", "-y", "-progress", "pipe:2", "-nostats"]

    # inputs
    for media in medias:
        input_path = os.path.join(Config.creator_folder(post.user), media.file_path)
        cmd.extend(["-i", input_path])

    # filter complex
    filter_lines = []
    for idx in range(len(medias)):
        filter_lines.append(
            f"[{idx}:v]"
            f"scale={width}:{height}:force_original_aspect_ratio=decrease,"
            f"pad={width}:{height}:(ow-iw)/2:(oh-ih)/2,"
            f"fps={Config.FPS},setsar=1"
            f"[v{idx}]"
        )

    # concat inputs
    concat = []
    for idx in range(len(medias)):
        concat.append(f"[v{idx}][{idx}:a]")

    filter_lines.append(f"{''.join(concat)}concat=n={len(medias)}:v=1:a=1[outv][outa]")
    filter_complex = ";".join(filter_lines)

    cmd.extend(
        [
            "-filter_complex",
            filter_complex,
            "-map",
            "[outv]",
            "-map",
            "[outa]",
            "-c:v",
            "libx264",
            "-preset",
            Config.PRESET,
            "-threads",
            str(Config.THREADS),
            "-c:a",
            "aac",
            "-movflags",
            "+faststart",
            output_path,
        ]
    )

    return {"cmd": cmd, "output_path": output_path}


def parse_line_ffmpeg_concat_into_advance(line: str) -> int | None:
    line = line.strip()
    if not line:
        return None

    progres_key = "out_time_ms"
    if line.startswith(progres_key):
        current_progress_str = line.replace(f"{progres_key}=", "").strip()
        try:
            current_progress_us = int(current_progress_str)
            current_progress_ms = current_progress_us // 1000
            return current_progress_ms
        except ValueError as e:
            logging.warning(
                "Skipping invalid progress line: %r (%s)",
                current_progress_str,
                e,
            )
            return None
        except Exception as e:
            UI.error(f"Unexpected error while updating progress: {e}")
            return None
    return None


def ffmpeg_concat(medias: list[Media], post: Post, progress: NestedProgress):
    """Run ffmpeg concat command to merge video together"""

    command_builder = ffmpeg_concat_build_command(medias, post)
    cmd = command_builder["cmd"]

    logging.info("CMD: %s", " ".join(cmd))

    ffmpeg_log = Config.CACHE_DIR / "ffmpeg.log"
    with open(ffmpeg_log, "w", encoding="utf-8") as log_file:
        print(cmd, file=log_file)
        # run cmd
        process = subprocess.Popen(
            cmd,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.PIPE,
            text=True,
            bufsize=1,
        )

        assert process.stderr is not None
        total_duration = get_total_duration(medias, post)
        progress.start_current(
            description=f"{post.user}->{medias[0].file_path}", total=total_duration
        )

        last_progress = 0

        for line in process.stderr:
            line = line.strip()
            print(line, file=log_file)
            current_progress_ms = parse_line_ffmpeg_concat_into_advance(line)
            if current_progress_ms is None:
                continue
            delta = current_progress_ms - last_progress
            progress.advance_current(step=delta)
            last_progress = current_progress_ms

        process.wait()
        progress.finish_current()

    UI.debug(f"Result: {process.returncode}")
    if process.returncode != 0:
        UI.error(f"Failed to concat videos. See ffmpeg log file {ffmpeg_log}")
        with open(ffmpeg_log, "r", encoding="utf-8") as f:
            lines = f.read()
        logging.warning("---FFMPEG LOG---")
        logging.warning(lines)
        logging.warning("---END FFMPEG LOG---")
        return process.returncode

    return 0


def parse_line_into_pourcent(line: str) -> float | None:
    line = line.strip()
    if not line:
        return None

    if "%" in line:
        try:
            parts = line.split("%")
            parts = parts[0].strip().split(" ")
            pourcent = parts[-1]
            flt_prcnt = float(pourcent)
            return flt_prcnt
        except Exception as e:
            UI.error(f"Error parsing line {line}: {e}")
            return None
    return None


def handbrake_optimized(media: Media, user: str, progress: NestedProgress):
    """Optimize video size with handbrake software"""

    handbrake_process = Config.HANDBRAKE_RUN_CMD.split(" ")

    folder_path = Config.creator_folder(user)
    video_path = os.path.join(folder_path, media.file_path)

    output_path = video_path + ".opti.mp4"

    cmd = ["-i", video_path, "-o", output_path, "--preset", "HQ 1080p30 Surround"]

    full_cmd = handbrake_process + cmd
    UI.debug(f"Running cmd '{full_cmd}'")

    # -- process
    process = subprocess.Popen(
        full_cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.DEVNULL,
        text=True,
    )

    assert process.stdout is not None
    progress.start_current(description="Optimizing", total=100)
    progress.set_status(
        f"{user}@({media.service}) -> ",
        f"{media.file_path} ({bytes_to_str(media.file_size)})",
    )

    current_progress = 0.0

    for line in process.stdout:
        float_pourcent = parse_line_into_pourcent(line)
        if float_pourcent is None:
            continue
        delta = float_pourcent - current_progress
        current_progress = float_pourcent
        progress.advance_current(step=delta)

    process.wait()
    progress.finish_current()
    # -- end process

    if process.returncode == 0:
        UI.debug("Return code: 0")
    else:
        UI.error(f"Return code: {process.returncode}")

    return process.returncode



# gui/db_viewer.py

import streamlit as st
import sqlite3

import pandas as pd

from rcdl.core.config import Config

TABLES = ["medias", "posts", "fuses"]


def get_table_columns(table_name):
    conn = sqlite3.connect(Config.DB_PATH)
    cur = conn.cursor()
    cur.execute(f"PRAGMA table_info({table_name})")
    columns = [info[1] for info in cur.fetchall()]
    conn.close()
    return columns


def get_table_data(table_name, sort_by=None, ascending=True):
    conn = sqlite3.connect(Config.DB_PATH)
    df = pd.read_sql_query(f"SELECT * FROM {table_name}", conn)
    conn.close()
    if sort_by and sort_by in df.columns:
        df = df.sort_values(by=sort_by, ascending=ascending)
    return df


def run_db_viewer():
    st.set_page_config(page_title="DB Viewer", layout="wide")
    st.title("Database Viewer")

    table_name = st.selectbox("Select Table", TABLES)

    # Load data
    df = get_table_data(table_name, sort_by=None, ascending=True)

    st.write(f"Showing `{table_name}` table ({len(df)} rows)")
    st.dataframe(df, width="stretch")



# core/db_queries.py
"""
Hold SQL STRING
"""

CREATE_MEDIAS_TABLE = """
CREATE TABLE IF NOT EXISTS medias (
    id INTEGER PRIMARY KEY AUTOINCREMENT,
    post_id TEXT,
    service TEXT,
    url TEXT,
    duration REAL,
    sequence INTEGER,
    status TEXT,
    checksum TEXT,
    file_path TEXT,
    created_at DATETIME,
    updated_at DATETIME,
    file_size INTEGER,
    fail_count INTEGER,
    UNIQUE(post_id, url)
)
"""

CREATE_POSTS_TABLE = """
CREATE TABLE IF NOT EXISTS posts (
    id TEXT PRIMARY KEY,
    user TEXT,
    service TEXT,
    domain TEXT,
    published DATETIME,
    json_hash TEXT,
    raw_json JSON,
    fetched_at DATETIME
)
"""

CREATE_FUSE_TABLE = """
CREATE TABLE IF NOT EXISTS fuses (
    id TEXT PRIMARY KEY,
    duration INTEGER,
    total_parts INTEGER,
    status TEXT,
    checksum TEXT,
    file_path TEXT,
    created_at DATETIME,
    updated_at DATETIME,
    file_size INTEGER,
    fail_count INTEGER
)
"""

INSERT_POST = """
INSERT OR IGNORE INTO posts (
    id, user, service, domain, published,
    json_hash, raw_json, fetched_at
)
VALUES (?, ?, ?, ?, ?, ?, ?, ?)
"""

INSERT_FUSED_MEDIA = """
INSERT OR IGNORE INTO fuses (
    id, duration, total_parts, status, checksum,
    file_path, created_at, updated_at, file_size, fail_count
)
VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
"""

INSERT_MEDIA = """
INSERT OR IGNORE INTO medias (
    post_id, service, url, duration, sequence, status,
    checksum, file_path, created_at, updated_at, file_size, fail_count
)
VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)
"""

UPDATE_MEDIA = """
UPDATE medias
SET duration = ?, file_size = ?, checksum = ?, status = ?,
    created_at = ?, updated_at = ?, fail_count = ?
WHERE post_id = ? AND url = ?
"""

UPDATE_FUSE = """
UPDATE fuses
SET duration = ?, status = ?, checksum = ?,
    created_at = ?, updated_at = ?, file_size = ?,
    fail_count = ?
WHERE id = ?
"""

QUERY_POST_ID = "SELECT * FROM posts WHERE id = ?"
QUERY_POST_USER = "SELECT * FROM posts WHERE user = ?"
QUERY_MEDIA_STATUS = "SELECT * FROM medias WHERE status = ?"
QUERY_MEDIA_ID = "SELECT * FROM medias WHERE post_id = ?"
QUERY_MEDIA_ID_AND_FILEPATH = "SELECT * FROM medias WHERE post_id = ? AND file_path = ?"
QUERY_FUSES_STATUS = "SELECT * FROM fuses WHERE status = ?"
QUERY_FUSES_ID = "SELECT * FROM fuses WHERE id = ?"



# gui/video_server.py

from fastapi import FastAPI, Query
from fastapi.responses import FileResponse
import os

app = FastAPI()


@app.get("/video")
async def get_video(path: str = Query(...)):
    real_path = path
    print("PATH FROM FASTAPI:", repr(real_path))

    if not os.path.isfile(real_path):
        return {"error": "File not found"}, 404

    return FileResponse(
        real_path,
        media_type="video/mp4",
        headers={
            "Accept-Ranges": "bytes",
        },
    )



# utils.py

"""Hold collection of useful functions"""

from datetime import datetime, timezone
import hashlib
import json
from pathlib import Path
import os
import time

from rcdl.core.config import Config
from rcdl.interface.ui import UI


def get_date_now() -> str:
    """Return an str of current datetime in isoformat"""
    return datetime.now(timezone.utc).isoformat()


def get_media_metadata(full_path: str) -> tuple[int, int, str]:
    """Get duration (if possible), file_size and checksum"""
    from rcdl.core.downloader_subprocess import ffprobe_get_duration

    path = Path(full_path)
    if not os.path.exists(path):
        UI.error(f"{path} path should exist but does not.")

    file_size = path.stat().st_size
    checksum = get_media_hash(path)
    if checksum is None:
        checksum = ""
    duration = ffprobe_get_duration(path)
    if duration is None:
        duration = 0
    return duration, file_size, checksum


def get_media_hash(path: Path, retries: int = Config.CHECKSUM_RETRY) -> str | None:
    """Return a hash of a video file"""
    for attempt in range(retries + 1):
        try:
            sha256_hash = hashlib.sha256()
            with path.open("rb") as f:
                for chunk in iter(lambda: f.read(8192), b""):
                    sha256_hash.update(chunk)
            checksum = sha256_hash.hexdigest()
            return checksum
        except OSError as e:
            if attempt >= retries:
                UI.error(
                    f"OSError: Failed to checksum {path} "
                    f"at attempt {attempt + 1} due to: {e}"
                )
                return None
            UI.warning(
                f"OSError: Failed to checksum {path} "
                f"at attempt {attempt + 1} due to: {e}"
            )
            time.sleep(1.0)


def get_json_hash(data: dict) -> tuple[str, str]:
    """Return a hash of a dict"""
    raw_json = json.dumps(data, sort_keys=True)
    json_hash = hashlib.sha256(raw_json.encode("utf-8")).hexdigest()
    return raw_json, json_hash


def bytes_to_mb(bytes: float | int) -> float:
    return bytes / (1024 * 1024)


def bytes_to_str(bytes: float | int) -> str:
    mb = bytes_to_mb(bytes)

    if mb > 1024:
        gb = mb / 1024
        return f"{gb:.1f} GB"
    return f"{mb:.1f} MB"


def clean_all(all: bool, partial: bool, cache: bool, medias_deleted: bool):
    """Remove partial file & external programm cache"""

    from rcdl.core.db import DB
    from rcdl.core.models import Status
    from rcdl.core.downloader_subprocess import ytdlp_clear_cache, kill_aria2c

    if all:
        partial = True
        cache = True
        medias_deleted = True

    # remove all partial file
    if partial:
        path = Config.BASE_DIR
        folders = os.listdir(path)

        for folder in folders:
            if folder.startswith("."):
                continue

            folder_full_path = os.path.join(path, folder)
            files = os.listdir(folder_full_path)

            for file in files:
                if (
                    file.endswith(".aria2")
                    or file.endswith(".part")
                    or file.endswith(".opti.mp4")
                ):
                    full_path = os.path.join(folder_full_path, file)
                    os.remove(full_path)
                    UI.info(f"Removed {full_path}")

    # cache
    if cache:
        # clear yt-dlp cache
        ytdlp_clear_cache()
        UI.info("Cleared yt-dlp cache dir")

        # kill aria2c
        kill_aria2c()
        UI.info("Kill aria2c")

    # medias with status deleted
    if medias_deleted:
        with DB() as db:
            medias = db.query_media_by_status(Status.TO_BE_DELETED)
        UI.info(f"Found {len(medias)} medias with DELETED status")

        if len(medias) == 0:
            return

        total_size = 0.0
        total_vid = 0
        for media in medias:
            with DB() as db:
                post = db.query_post_by_id(media.post_id)
            if post is None:
                UI.warning(f"Could not match post id {media.post_id} to a post")
                continue

            path = os.path.join(Config.creator_folder(post.user), media.file_path)
            if not os.path.exists:
                UI.error(f"Path {path} should exist. Does not")
                continue

            try:
                os.remove(path)
                UI.info(f"Removed '{path}' ({bytes_to_mb(media.file_size)} MB)")
                total_size += media.file_size
                total_vid += 1
            except Exception as e:
                UI.error(f"Failed to rm media {media.post_id}/{media.url} due to: {e}")
                continue

            with DB() as db:
                media.status = Status.DELETED
                db.update_media(media)

        UI.info(
            f"Removed a total of {total_vid} videos ({bytes_to_mb(total_size)}MB,"
            f" {round(bytes_to_mb(total_size) / 100, 1)}GB)"
        )


def format_seconds(seconds: int | float) -> str:
    seconds = int(seconds)
    h = seconds // 3600
    m = (seconds % 3600) // 60
    s = seconds % 60

    if h > 0:
        return f"{h}h:{m:02d}m:{s:02d}s"
    elif m > 0:
        return f"{m}m:{s:02d}s"
    else:
        return f"{s}s"



# interface/cli.py

"""Hold all cli commands"""

import logging
import subprocess
import sys
import inspect

import click

from rcdl.core import downloader as dl
from rcdl.interface.ui import UI, print_db_info
from rcdl.core.fuse import fuse_medias
from rcdl.core.opti import optimize
from rcdl.core.config import Config
from rcdl.core.parser import (
    get_creators,
    get_creator_from_line,
    parse_creator_input,
    append_creator,
)
from rcdl.core.db import DB
from rcdl.utils import clean_all

from rcdl import __version__


@click.command(help="Refresh video to be downloaded")
def refresh():
    """Refresh database with creators videos

    - get all creators from creators.txt
    - for each creators find all videos and put them in the database
    No download is done in this command
    """

    UI.info("Welcome to RCDL refresh")
    dl.refresh_creators_videos()

    with DB() as db:
        info = db.get_nb_per_status()
    print_db_info(info)


@click.command(help="Download all videos from all creator")
@click.option(
    "--max-fail-count",
    type=int,
    help="Set maximum number of failed attempts. Take precedence over config.toml",
)
def dlsf(max_fail_count: int | None):
    """Download video based on DB information

    - read databse
    - for each video NOT_DOWNLOADED or FAILED & fail_count < settings, dl video
    """
    UI.info("Welcome to RCDL dlsf")
    dl.download_videos_to_be_dl(max_fail_count)


@click.command("fuse", help="Fuse part video into one")
def fuse():
    """Fuse videos"""
    UI.info("Fuse/Concat video together")
    fuse_medias()


@click.command(help="Optimized video size")
def opti():
    """Optimized video size"""
    UI.info("Optimized video")
    optimize()


@click.command(help="Clean patial download file, cache, etc...")
@click.option("--all", is_flag=True, help="Act as if al lthe flags are true")
@click.option(
    "--partial", is_flag=True, help="Remove partial file from download (.aria2, .part)"
)
@click.option(
    "--cache", is_flag=True, help="remove cache from yt-dlp and kill aria2 process"
)
@click.option(
    "--medias-deleted",
    is_flag=True,
    help="Remove media marked for deletion (TO_BE_DELETED status)",
)
def clean(all: bool, partial: bool, cache: bool, medias_deleted: bool):
    """Remove partial download, clear subprocesses cache"""
    clean_all(all, partial, cache, medias_deleted)


@click.command(help="Discover videos/creators with tags")
@click.option("--tag", required=True, type=str, help="Tag to search for")
@click.option(
    "--max-page", default=10, type=int, help="Maximum number of pages to fetch"
)
def discover(tag, max_page):
    """Discover new creators/videos
    currently WIP. Do not use in prod"""
    msg = f"[cdl] discover with tag={tag} max_page={max_page}"
    click.echo(msg)
    logging.info(msg)
    UI.info("WIP - UNIMPLEMENTED")


@click.command("add", help="Add a creator")
@click.argument("creator_input")
def add_creator(creator_input):
    """Add a creator (URL or str) to creators.txt"""
    service, creator_id = parse_creator_input(creator_input)
    line = f"{service}/{creator_id}"
    creator = get_creator_from_line(line)
    if creator is not None:
        append_creator(creator)
        UI.info(f"Added {line} to creators.txt")
    else:
        UI.warning("Could not extract creator from input. Please check input is valid")


@click.command("remove", help="Remove a creator")
@click.option("--db", is_flag=True)
@click.argument("creator_input")
def remove_creator(db, creator_input):
    """Remove a creator (excat line) from creators.txt"""
    _service, creator_id = parse_creator_input(str(creator_input))

    creators = get_creators()
    all_creators = []
    matched_creator = None
    for creator in creators:
        if creator.id == creator_id:
            matched_creator = creator
            continue
        all_creators.append(creator)

    if matched_creator is None:
        UI.error(f"Could not find creator from {creator_input}")
        return

    try:
        with open(Config.CREATORS_FILE, "w", encoding="utf-8"):
            pass
    except OSError as e:
        UI.error(
            f"Failed to create creators file at {Config.CREATORS_FILE} due to: {e}"
        )
        return
    for c in all_creators:
        append_creator(c)
    UI.info(f"Removed creator {matched_creator.id}@({matched_creator.service})")
    if db:
        UI.info("Not yet implemented")


@click.command("list", help="List all creators")
def list_creators():
    """List all creators in creators.txt"""
    creators = get_creators()
    UI.table_creators(creators)


@click.command("status", help="Print db info")
def db_status():
    """Print number of entry per status per tables in the database"""
    with DB() as db:
        info = db.get_nb_per_status()
    print_db_info(info)


@click.command("gui", help="Launch GUI")
def launch_gui():
    UI.info("Launching GUI...")
    import importlib.resources as resources

    gui_path = resources.files("rcdl.gui").joinpath("gui.py")
    subprocess.Popen(
        [
            sys.executable,
            "-m",
            "uvicorn",
            "rcdl.gui.video_server:app",
            "--host",
            "0.0.0.0",
            "--port",
            "8123",
            "--log-level",
            "warning",
        ]
    )
    subprocess.run([sys.executable, "-m", "streamlit", "run", str(gui_path)])


@click.command("show-config")
def show_config():
    """Show all value in Config"""
    for name, value in vars(Config).items():
        if not name.startswith("__") and not inspect.isroutine(value):
            UI.success(f"{name}: {value}")


# --- CLI GROUP ---
@click.group()
@click.option("--debug", is_flag=True)
@click.option("--dry-run", is_flag=True)
@click.version_option(version=__version__, prog_name=Config.APP_NAME)
def cli(debug, dry_run):
    """Init cli app. Assign Config var depending on flag used when calling prgm"""
    Config.set_debug(debug)
    Config.set_dry_run(dry_run)


# main commands
cli.add_command(dlsf)
cli.add_command(discover)
cli.add_command(refresh)
cli.add_command(fuse)
cli.add_command(opti)
cli.add_command(launch_gui)

# creators command
cli.add_command(add_creator)
cli.add_command(remove_creator)
cli.add_command(list_creators)

# helper command
cli.add_command(clean)
cli.add_command(db_status)
cli.add_command(show_config)
